import React, { useCallback, useEffect, useState } from 'react';
import Calendar from '../Calendar';
import { findNextRangeIndex, generateStyles, isSelectionComplete } from '../../utils';
import { isBefore, differenceInCalendarDays, addDays, min, isWithinInterval, max, startOfDay, endOfDay } from 'date-fns';
import coreStyles, { ClassNames, DEFAULT_COLOR } from '../../styles';
import { DateRangeProps, isNoEndDateRange, isNotFullyEmptyRange, isSureRange, NotFullyEmptyRange, PreviewRange, MaybeEmptyRange, RangeFocus, SureStartSureEndDate, ClearPreview, IgnorePreview, IGNORE_PREVIEW, CLEAR_PREVIEW } from '../../types';
import classnames from 'classnames';
import { compose } from 'ramda';

type DefaultComponentProps = {
  classNames?: Partial<ClassNames>;
  ranges?: NotFullyEmptyRange[];
  moveRangeOnFirstSelection?: boolean;
  retainEndDateOnFirstSelection?: boolean;
  rangeColors?: string[];
  disabledDates?: Date[],
};
export type DateRangeDCP = DefaultComponentProps;

type ComponentProps = DateRangeProps & DefaultComponentProps & {
  onRangeFocusChange?: (focusedRange: RangeFocus) => void;
};
export type DateRangeCP = ComponentProps;

export type RangeSelection = {
  wasValid: boolean;
  range: {
    startDate: Date;
    endDate: Date | null;
  };
  nextFocusRange: RangeFocus;
};


const calculateEndDate = (value: Date, now: Date, endDate: Date | null, dayOffset: number, moveRangeOnFirstSelection: boolean, retainEndDateOnFirstSelection: boolean, maxDate?: Date) => {
  const userEndDate = (() => {
    if (moveRangeOnFirstSelection) {
      return addDays(value, dayOffset);
    }
    if (retainEndDateOnFirstSelection) {
      if (!endDate || isBefore(value, endDate)) {
        return endDate;
      }
      return value;
    }
    return value || now;
  })();
  const nextEndDate = maxDate ? (userEndDate ? min([userEndDate, maxDate]) : maxDate) : userEndDate ;
  // if (nextEndDate === null) {
  //   throw new Error('Bug! endDate is null, and it should not. RetainEndDateOnFirstSelection tiggers buggy behavior, what should the endDate be now that retainEDFS is true?');
  // }
  return nextEndDate;
};

type ShapeChangingParams = { value: Date | MaybeEmptyRange; };
type BaseParams = { focusedRange: RangeFocus; disabledDates: Date[]; ranges: MaybeEmptyRange[]; }
type ComputeProps = BaseParams & ShapeChangingParams & { selectedRange: NotFullyEmptyRange; moveRangeOnFirstSelection: boolean; retainEndDateOnFirstSelection: boolean; maxDate?: Date; };
function computeStartDateEndDate({ value, selectedRange, focusedRange, disabledDates, moveRangeOnFirstSelection, retainEndDateOnFirstSelection, ranges, maxDate }: ComputeProps) {
  const base = {
    isStartDateSelected: focusedRange[1] === 0,
    focusedRange,
    disabledDates,
    ranges,
  }
  if (isNotFullyEmptyRange(value)) {
    if (!isSureRange(value)) {
      console.log(value);
      throw new Error('Bug, expecting value to be a SureRange (aka startDate and endDate instanceof Date), but not the case!');
    }
    return {
      ...base,
      startDate: value.startDate,
      endDate: value.endDate,
    };
  }
  if (!(value instanceof Date)) {
    console.log(value);
    throw new Error('Probably got a fully empty range, expecting either Date or NotFullyEmptyRange');
  }
  if (isSelectionComplete(focusedRange)) {
    // startDate selection
    const now = new Date();
    const dayOffset = differenceInCalendarDays(selectedRange.endDate || now, selectedRange.startDate);
    return {
      ...base,
      startDate: value,
      endDate: calculateEndDate(value, now, selectedRange.endDate, dayOffset, moveRangeOnFirstSelection, retainEndDateOnFirstSelection, maxDate),
      nextFocusRange: [focusedRange[0], 1] as RangeFocus,
    };
  } else {
    return {
      ...base,
      startDate: selectedRange.startDate,
      endDate: value,
    }
  }
}

type FlipProps = BaseParams & NotFullyEmptyRange & { nextFocusRange?: RangeFocus; isStartDateSelected: boolean; };
const flipIfReversed = (params: FlipProps) => {
  const { startDate, endDate, isStartDateSelected } = params;
  return endDate && isBefore(endDate, startDate)
    ? {
      ...params,
      isStartDateSelected: !isStartDateSelected,
      startDate: startOfDay(endDate),
      endDate: endOfDay(startDate),
    }
    : {
      ...params,
      startDate: startOfDay(startDate),
      endDate: endDate && endOfDay(endDate)
    };
}

const getNextFocusRange = (ranges: MaybeEmptyRange[], focusedRange: RangeFocus, nextFocusRange?: RangeFocus): RangeFocus => {
  if (nextFocusRange) return nextFocusRange;
  const nextFocusRangeIndex = findNextRangeIndex(ranges, focusedRange[0]);
  return [nextFocusRangeIndex, 0];
}

const computeRange = ({ disabledDates, startDate, endDate, isStartDateSelected, focusedRange, nextFocusRange, ranges }: FlipProps) => {
  const inValidDatesWithinRange = disabledDates.filter(disabledDate =>
    endDate && isWithinInterval(disabledDate, {
      start: startDate,
      end: endDate,
  }));

  const wasValid = !(inValidDatesWithinRange.length > 0);

  const range = {
    startDate: !wasValid && isStartDateSelected
      ? addDays(max(inValidDatesWithinRange), 1)
      : startDate,
    endDate: !wasValid && !isStartDateSelected
      ? addDays(min(inValidDatesWithinRange), -1)
      : endDate,
  };

  return {
    wasValid,
    range,
    nextFocusRange: getNextFocusRange(ranges, focusedRange, nextFocusRange),
  };
}

const getRes = compose(computeRange, flipIfReversed, computeStartDateEndDate);

export const calcNewSelectionGen = (props: Readonly<ComponentProps>, stateFocusedRange: RangeFocus) => (value: Date | MaybeEmptyRange | IgnorePreview | ClearPreview): RangeSelection | IgnorePreview | ClearPreview => {

  if (value === IGNORE_PREVIEW || value === CLEAR_PREVIEW) return value;

  const focusedRange = props.focusedRange || stateFocusedRange;
  const {
    ranges,
    onChange,
    maxDate,
    moveRangeOnFirstSelection,
    retainEndDateOnFirstSelection,
    disabledDates,
  } = props;
  const focusedRangeIndex = focusedRange[0];
  const selectedRange = ranges && ranges[focusedRangeIndex];

  if (!selectedRange || !onChange) return IGNORE_PREVIEW;

  if (!isSureRange(selectedRange) && !isNoEndDateRange(selectedRange)) {
    console.log('selectedRange', selectedRange);
    console.log('value', value);
    throw new Error('Bug, expecting selected range to be a sure range or no end date range, but it is neither');
  }

  return getRes({
    value,
    selectedRange,
    focusedRange,
    disabledDates: disabledDates || [],
    moveRangeOnFirstSelection: moveRangeOnFirstSelection || false,
    ranges,
    retainEndDateOnFirstSelection: retainEndDateOnFirstSelection || false,
    maxDate,
  });

};

const setSelectionGen = (props: ComponentProps, stateFocusedRange: RangeFocus, setFocusedRange: React.Dispatch<React.SetStateAction<RangeFocus>>, setPreview: React.Dispatch<React.SetStateAction<PreviewRange | ClearPreview | IgnorePreview>>) => (value: Date | SureStartSureEndDate<Date>) => {
  const { onChange, ranges, onRangeFocusChange } = props;
  const focusedRange = props.focusedRange || stateFocusedRange;
  const focusedRangeIndex = focusedRange[0];
  const selectedRange = ranges && ranges[focusedRangeIndex];
  if (!selectedRange) return;
  const newSelection = calcNewSelectionGen(props, focusedRange)(value);
  if (!newSelection) {
    throw new Error('Bug, expecting new selection to not be undefined');
  }
  const isSelectionComplete = newSelection.nextFocusRange[1] === 0;
  onChange && onChange({
    [selectedRange.key || `range${focusedRangeIndex + 1}`]: {
      ...selectedRange,
      ...newSelection.range,
    },
  }, isSelectionComplete);

  setFocusedRange(newSelection.nextFocusRange);
  setPreview(null);
  onRangeFocusChange && onRangeFocusChange(newSelection.nextFocusRange);
};

const handleRangeFocusChangeGen = (setFocusedRange: React.Dispatch<React.SetStateAction<RangeFocus>>, props: ComponentProps) => (focusedRange: RangeFocus) => {
  setFocusedRange(focusedRange);
  props.onRangeFocusChange && props.onRangeFocusChange(focusedRange);
};
const updatePreviewGen = (setPreview: React.Dispatch<React.SetStateAction<PreviewRange | ClearPreview | IgnorePreview>>, props: ComponentProps, stateFocusedRange: RangeFocus) => (newSelection?: RangeSelection | IgnorePreview | ClearPreview) => {
  if (newSelection === IGNORE_PREVIEW) return;
  if (newSelection === CLEAR_PREVIEW) {
    setPreview(null);
    return;
  }
  const { rangeColors, ranges } = props;
  const focusedRange = props.focusedRange || stateFocusedRange;
  const color = (ranges && ranges[focusedRange[0]]?.color) || (rangeColors && rangeColors[focusedRange[0]]);
  setPreview({ ...newSelection.range, color });
};

const composedUpdatePreview = (setPreview: React.Dispatch<React.SetStateAction<PreviewRange | ClearPreview | IgnorePreview>>, props: ComponentProps, stateFocusedRange: RangeFocus) => (nextPreviewRangeOrDate: Date | MaybeEmptyRange | IgnorePreview | ClearPreview) => {
  if (nextPreviewRangeOrDate === IGNORE_PREVIEW) return;
  const calcNewSelection = calcNewSelectionGen(props, stateFocusedRange);
  const updatePreview = updatePreviewGen(setPreview, props, stateFocusedRange);
  compose(updatePreview, calcNewSelection)(nextPreviewRangeOrDate);
}

export const defaultProps: DefaultComponentProps = {
  classNames: {},
  ranges: [],
  moveRangeOnFirstSelection: false,
  retainEndDateOnFirstSelection: false,
  rangeColors: [DEFAULT_COLOR, '#3ecf8e', '#fed14c'],
  disabledDates: [],
}

export function getInitialFocusedRange(initialFocusedRange: ComponentProps["initialFocusedRange"], ranges: ComponentProps["ranges"]) {
  return initialFocusedRange || [findNextRangeIndex(ranges || []), 0];
}

const DateRange: React.FC<ComponentProps> = props => {

  const allProps: ComponentProps = { ...defaultProps, ...props };
  const { initialFocusedRange, ranges, classNames, className, previewRange: nextPreviewRange } = allProps;
  const [focusedRange, setFocusedRange] = useState<RangeFocus>(getInitialFocusedRange(initialFocusedRange, ranges));
  const [previewRange, setPreviewRange] = useState<PreviewRange | IgnorePreview | ClearPreview>(CLEAR_PREVIEW);
  const styles = generateStyles([coreStyles, classNames]);
  const setSelection = setSelectionGen(allProps, focusedRange, setFocusedRange, setPreviewRange);
  const handleRangeFocusChange = handleRangeFocusChangeGen(setFocusedRange, allProps);
  const fullUpdatePreview = useCallback(
    composedUpdatePreview(setPreviewRange, props, focusedRange),
    [setPreviewRange, props, focusedRange],
  ) ;

  useEffect(() => {
    if (previewRange === nextPreviewRange) return;
    fullUpdatePreview(nextPreviewRange);
  }, [fullUpdatePreview, nextPreviewRange])

  return (
    <Calendar
      {...allProps}
      focusedRange={focusedRange}
      onRangeFocusChange={handleRangeFocusChange}
      previewRange={previewRange}
      onPreviewChange={fullUpdatePreview}
      displayMode={"dateRange"}
      className={classnames(styles.dateRangeWrapper, className)}
      onChange={setSelection}
      updateRange={setSelection}
    />
  );

}

export default DateRange;
