/* eslint-disable no-fallthrough */
import React, { FocusEvent, KeyboardEvent, MouseEvent, useCallback, useState } from 'react';
import classnames from 'classnames';
import { format } from 'date-fns';
import { PreviewRange, MaybeEmptyRange, DEFAULT_DISPLAY_MODE, ClearPreview, CLEAR_PREVIEW } from '../../types';
import { CoreStyles } from '../../styles';
import Preview from './Preview';
import Selection from './Selection';

export type InRange = ({
  isStartEdge: boolean;
  isEndEdge: boolean;
  isInRange: boolean;
} & MaybeEmptyRange);

export type DateReceivingFunc = (date: Date) => void;
export type DateOrClearReceivingFunc = (date: Date | ClearPreview) => void;

type ComponentProps = SomeComponentProps & {
  color?: string;
  date?: Date;
  day: Date;
  dayContentRenderer?: DateReceivingFunc;
  dayDisplayFormat: string;
  displayMode?: 'dateRange' | 'date';
  isSelectionComplete: boolean;
  onMouseDown: DateReceivingFunc;
  onMouseEnter: DateReceivingFunc;
  onMouseUp: DateReceivingFunc;
  onPreviewChange?: DateOrClearReceivingFunc;
  previewRange?: PreviewRange | ClearPreview;
  previewColor?: string;
  ranges: MaybeEmptyRange[],
};

type SomeComponentProps = {
  disabled: boolean; // Whether the date is not selectable
  isEndOfMonth: boolean;
  isEndOfWeek: boolean;
  isPassive: boolean;
  isStartOfMonth: boolean;
  isStartOfWeek: boolean;
  isToday: boolean;
  isWeekend: boolean;
  styles: CoreStyles;
};

const getClassNames = ({
    disabled,
    isEndOfMonth,
    isEndOfWeek,
    isPassive,
    isStartOfMonth,
    isStartOfWeek,
    isToday,
    isWeekend,
    styles,
    hover,
    active,
  }: SomeComponentProps & { hover: boolean; active: boolean; }) => {
  return classnames(styles.day, {
    [styles.dayPassive]: isPassive,
    [styles.dayDisabled]: disabled,
    [styles.dayToday]: isToday,
    [styles.dayWeekend]: isWeekend,
    [styles.dayStartOfWeek]: isStartOfWeek,
    [styles.dayEndOfWeek]: isEndOfWeek,
    [styles.dayStartOfMonth]: isStartOfMonth,
    [styles.dayEndOfMonth]: isEndOfMonth,
    [styles.dayHovered]: hover,
    [styles.dayActive]: active,
  });
};


const handleKeyEventGen = (props: ComponentProps) => (event: KeyboardEvent<HTMLButtonElement>) => {
  const { day, onMouseDown, onMouseUp } = props;
  if ([13 /* space */, 32 /* enter */].includes(event.keyCode)) {
    if (event.type === 'keydown') onMouseDown(day);
    else onMouseUp(day);
  }
}

const handleMouseEventGen = (props: ComponentProps, setHover: React.Dispatch<React.SetStateAction<boolean>>, setActive: React.Dispatch<React.SetStateAction<boolean>>) => (event: MouseEvent<HTMLButtonElement> | FocusEvent<HTMLButtonElement>) => {
  const { day, disabled, onPreviewChange, onMouseEnter, onMouseDown, onMouseUp } = props;
  if (disabled) {
    onPreviewChange && onPreviewChange(CLEAR_PREVIEW);
    return;
  }

  switch (event.type) {
    case 'mouseenter':
      onMouseEnter(day);
      onPreviewChange && onPreviewChange(day);
      setHover(true);
      break;
    case 'blur':
    case 'mouseleave':
      setHover(false);
      break;
    case 'mousedown':
      setActive(true);
      onMouseDown(day);
      break;
    case 'mouseup':
      event.stopPropagation();
      setActive(false);
      onMouseUp(day);
      break;
    case 'focus':
      onPreviewChange && onPreviewChange(day);
      break;
  }
};

const DayCell: React.FC<ComponentProps> = (props) => {

  const [hover, setHover] = useState<boolean>(false);
  const [active, setActive] = useState(false);
  const { dayContentRenderer, previewRange, day, styles, ranges, date, displayMode, color, isSelectionComplete } = props;
  const handleMouseEvent = useCallback(
    handleMouseEventGen(props, setHover, setActive),
    [props, setHover, setActive],
  );
  const handleKeyEvent = useCallback(
    handleKeyEventGen(props),
    [props],
  );

  return (
    <button
      type="button"
      onMouseEnter={handleMouseEvent}
      onMouseLeave={handleMouseEvent}
      onFocus={handleMouseEvent}
      onMouseDown={handleMouseEvent}
      onMouseUp={handleMouseEvent}
      onBlur={handleMouseEvent}
      onPauseCapture={handleMouseEvent}
      onKeyDown={handleKeyEvent}
      onKeyUp={handleKeyEvent}
      className={getClassNames({...props, hover, active })}
      {...(props.disabled || props.isPassive ? { tabIndex: -1 } : {})}
      style={{ color: props.color }}
    >
      <Selection
        isComplete={isSelectionComplete}
        styles={styles}
        ranges={ranges}
        day={day}
        displayMode={displayMode || DEFAULT_DISPLAY_MODE}
        date={date}
        color={color}
      />
      <Preview
        previewRange={previewRange}
        day={day}
        styles={styles}
      />
      <span className={props.styles.dayNumber}>
        {
          (dayContentRenderer && dayContentRenderer(props.day))
          || <span>{format(props.day, props.dayDisplayFormat)}</span>
        }
      </span>
    </button>
  );
}

export default DayCell;
